from io import BytesIO
from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    FileFormat,
    FileQuery,
    FileUploadRequest,
    KYCFileType,
)
from pydantic import HttpUrl
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Downloadable, Queryable, Uploadable


@dataclass
class File(Downloadable, Queryable, Uploadable):
    _resource: ClassVar = 'files'
    _query_params: ClassVar = FileQuery

    extension: str
    type: KYCFileType
    url: HttpUrl
    user_id: str

    @classmethod
    def upload(
        cls,
        file: BytesIO,
        file_type: KYCFileType,
        extension: Optional[str],
        is_back: bool = False,
        user_id: str = 'me',
        *,
        session: Session = global_session,
    ) -> 'File':
        """
        Stores an encrypted version of the file,
        only users with permissions can access it.

        :param file:
        :param user_id:
        :param session:
        :return: New encrypted file object
        """
        req = FileUploadRequest(
            file=file.read(),
            type=file_type,
            extension=extension,
            is_back=is_back,
            user_id=user_id,
        )
        return cast(
            'File',
            cls._upload(
                session=session,
                **req.dict(),
            ),
        )

    @property
    def file(self) -> bytes:
        """
        Bytes of the decrypted file.
        Format of the file is found on `file_type` property.
        """
        return self.download(self, file_format=FileFormat.any).read()

    @property
    def pdf(self) -> bytes:
        """
        Override from `Downloadable`, this property does not apply.
        """
        raise NotImplementedError

    @property
    def xml(self) -> bytes:
        """
        Override from `Downloadable`, this property does not apply.
        """
        raise NotImplementedError



import pytest
from cuenca_validations.types import EventType

from cuenca import UserEvent


@pytest.mark.vcr
def test_user_event_retrieve():
    user_id = 'USCM-zlFcNQk6ue4gZ_mTGeQ'
    user_event = UserEvent.one(user_id=user_id, type=EventType.created)
    assert user_event.id is not None



__all__ = [
    'ApiKey',
    'Account',
    'Arpc',
    'BalanceEntry',
    'BillPayment',
    'Card',
    'CardActivation',
    'CardTransaction',
    'CardValidation',
    'Commission',
    'CurpValidation',
    'Deposit',
    'Endpoint',
    'File',
    'FileBatch',
    'Identity',
    'IdentityEvent',
    'LoginToken',
    'Saving',
    'ServiceProvider',
    'Session',
    'Statement',
    'Transfer',
    'User',
    'UserEvent',
    'UserLogin',
    'WalletTransaction',
    'Webhook',
    'WhatsappTransfer',
]

from .accounts import Account
from .api_keys import ApiKey
from .arpc import Arpc
from .balance_entries import BalanceEntry
from .bill_payments import BillPayment
from .card_activations import CardActivation
from .card_transactions import CardTransaction
from .card_validations import CardValidation
from .cards import Card
from .commissions import Commission
from .curp_validations import CurpValidation
from .deposits import Deposit
from .endpoints import Endpoint
from .file_batches import FileBatch
from .files import File
from .identities import Identity
from .identity_events import IdentityEvent
from .login_tokens import LoginToken
from .resources import RESOURCES
from .savings import Saving
from .service_providers import ServiceProvider
from .sessions import Session
from .statements import Statement
from .transfers import Transfer
from .user_credentials import UserCredential
from .user_events import UserEvent
from .user_logins import UserLogin
from .users import User
from .wallet_transactions import WalletTransaction
from .webhooks import Webhook
from .whatsapp_transfers import WhatsappTransfer

# avoid circular imports
resource_classes = [
    ApiKey,
    Account,
    Arpc,
    BalanceEntry,
    BillPayment,
    Card,
    CardActivation,
    CardTransaction,
    CardValidation,
    CurpValidation,
    Commission,
    Deposit,
    Endpoint,
    File,
    FileBatch,
    Identity,
    IdentityEvent,
    LoginToken,
    Saving,
    Session,
    ServiceProvider,
    Statement,
    Transfer,
    User,
    UserCredential,
    UserEvent,
    UserLogin,
    WalletTransaction,
    WhatsappTransfer,
    Webhook,
]
for resource_cls in resource_classes:
    RESOURCES[resource_cls._resource] = resource_cls  # type: ignore



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types.requests import ARPCRequest
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Creatable


@dataclass
class Arpc(Creatable):
    """
    An ARPC (Authorisation Response Cryptogram) is generated by the issuer
    to authorize EMV transactions (Chip, Contactless)

    The point of sales terminal or ATM generate an ARQC (Authorization Request
    Cryptogram) to obtain authorization for transactions.
    After, the issuer has to verified the ARQC and generate an ARPC.
    Finally the ARPC is sent back to the point of sales terminal to authorize
    the transaction and validate the issuer as authentic
    """

    _resource: ClassVar = 'arpc'

    created_at: dt.datetime
    card_uri: str
    is_valid_arqc: Optional[bool]
    arpc: Optional[str]

    @classmethod
    def create(
        cls,
        number: str,
        arqc: str,
        arpc_method: str,
        transaction_data: str,
        response_code: str,
        transaction_counter: str,
        pan_sequence: str,
        unique_number: str,
        track_data_method: str,
        *,
        session: Session = global_session,
    ) -> 'Arpc':
        req = ARPCRequest(
            number=number,
            arqc=arqc,
            arpc_method=arpc_method,
            transaction_data=transaction_data,
            response_code=response_code,
            transaction_counter=transaction_counter,
            pan_sequence=pan_sequence,
            unique_number=unique_number,
            track_data_method=track_data_method,
        )
        return cast('Arpc', cls._create(session=session, **req.dict()))



from typing import ClassVar, Optional, cast

from cuenca_validations.types import DepositNetwork, DepositQuery
from pydantic.dataclasses import dataclass

from .accounts import Account
from .base import Transaction
from .resources import retrieve_uri


@dataclass
class Deposit(Transaction):
    _resource: ClassVar = 'deposits'
    _query_params: ClassVar = DepositQuery

    network: DepositNetwork
    source_uri: str
    tracking_key: Optional[str]  # clave rastreo if network is SPEI

    @property  # type: ignore
    def source(self) -> Account:
        return cast(Account, retrieve_uri(self.source_uri))



from typing import ClassVar

from cuenca_validations.types import StatementQuery
from pydantic.dataclasses import dataclass

from .base import Downloadable, Queryable


@dataclass
class Statement(Queryable, Downloadable):
    _resource: ClassVar = 'statements'
    _query_params: ClassVar = StatementQuery

    month: int
    year: int



import pytest

from cuenca import Endpoint
from cuenca.exc import CuencaResponseException


@pytest.mark.vcr
def test_endpoint_retrieve():
    id_endpoint = 'EN01'
    endpoint: Endpoint = Endpoint.retrieve(id_endpoint)
    assert endpoint.id == id_endpoint
    assert endpoint.url
    assert endpoint.events
    assert endpoint.is_active


@pytest.mark.vcr
def test_endpoint_create():
    endpoint: Endpoint = Endpoint.create(url='https://url.com')
    assert endpoint.id
    assert endpoint.events
    assert endpoint.is_enable
    assert endpoint.url
    assert endpoint.is_active


@pytest.mark.vcr
def test_endpoint_create_another():
    with pytest.raises(CuencaResponseException) as exc:
        Endpoint.create(url='https://url.xyz')
        assert exc.value
        assert exc.json['error'] == 'Usuario ya tiene webhook asociado'


@pytest.mark.vcr
def test_endpoint_update():
    id_endpoint = 'EN02'
    endpoint: Endpoint = Endpoint.update(
        endpoint_id=id_endpoint, url='https://url.io', is_enable=False
    )
    assert endpoint.id == id_endpoint
    assert endpoint.events
    assert endpoint.url == 'https://url.io'
    assert not endpoint.is_enable
    assert endpoint.is_active


@pytest.mark.vcr
def test_endpoint_deactivate():
    id_endpoint = 'EN02'
    endpoint: Endpoint = Endpoint.deactivate(id_endpoint)
    assert endpoint.id == id_endpoint
    assert not endpoint.is_active



import pytest

from cuenca.resources import CurpValidation, User


@pytest.mark.vcr
def test_user_create(user_request, curp_validation_request):
    # we need to create the curp validation first
    curp_validation = CurpValidation.create(**curp_validation_request)
    assert curp_validation.renapo_curp_match
    # creating the user
    user = User.create(**user_request)
    assert user.id is not None
    assert user.level == 0


@pytest.mark.vcr
def test_user_query():
    user = User.one(email_address='jose@test.com')
    assert user.id is not None


@pytest.mark.vcr
def test_user_update():
    user_id = 'USCM-zlFcNQk6ue4gZ_mTGeQ'
    changes = dict(
        profession='programmer',
        phone_number='+525555555555',
        govt_id=dict(
            type='ine',
            uri_front='cuenca.com',
            uri_back='cuenca.com',
            is_mx=True,
        ),
    )
    user = User.update(user_id, **changes)
    assert all(item in user.to_dict().keys() for item in changes.keys())


@pytest.mark.vcr
def test_user_identity_retrieve():
    user_id = 'USCM-zlFcNQk6ue4gZ_mTGeQ'
    user = User.retrieve(user_id)
    assert user_id == user.id
    identity = user.identity
    assert identity.id is not None



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import CardStatus, CardType
from cuenca_validations.types.requests import CardValidationRequest
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Creatable
from .cards import Card
from .resources import retrieve_uri


@dataclass
class CardValidation(Creatable):
    _resource: ClassVar = 'card_validations'

    created_at: dt.datetime
    card_uri: str
    user_id: str
    card_status: CardStatus
    card_type: CardType
    is_valid_cvv: Optional[bool]
    is_valid_cvv2: Optional[bool]
    is_valid_icvv: Optional[bool]
    is_valid_pin_block: Optional[bool]
    is_valid_exp_date: Optional[bool]
    is_pin_attempts_exceeded: bool
    is_expired: bool

    @classmethod
    def create(
        cls,
        number: str,
        cvv: Optional[str] = None,
        cvv2: Optional[str] = None,
        icvv: Optional[str] = None,
        exp_month: Optional[int] = None,
        exp_year: Optional[int] = None,
        pin_block: Optional[str] = None,
        pin_attempts_exceeded: Optional[str] = None,
        *,
        session: Session = global_session,
    ) -> 'CardValidation':
        req = CardValidationRequest(
            number=number,
            cvv=cvv,
            cvv2=cvv2,
            icvv=icvv,
            exp_month=exp_month,
            exp_year=exp_year,
            pin_block=pin_block,
            pin_attempts_exceeded=pin_attempts_exceeded,
        )
        return cast(
            'CardValidation', cls._create(session=session, **req.dict())
        )

    @property
    def card(self) -> Card:
        return cast(Card, retrieve_uri(self.card_uri))

    @property
    def card_id(self) -> str:
        return self.card_uri.split('/')[-1]

    @property
    def is_active(self):
        return self.card_status == CardStatus.active



import datetime as dt
from typing import ClassVar, Optional, cast

from cuenca_validations.types import (
    Country,
    CurpValidationRequest,
    Gender,
    State,
)
from cuenca_validations.types.identities import CurpField
from pydantic.dataclasses import dataclass

from ..http import Session, session as global_session
from .base import Creatable, Retrievable


@dataclass
class CurpValidation(Creatable, Retrievable):
    _resource: ClassVar = 'curp_validations'

    created_at: dt.datetime
    names: str
    first_surname: str
    second_surname: Optional[str]
    date_of_birth: dt.date
    country_of_birth: Country
    state_of_birth: State
    gender: Gender
    nationality: Optional[Country]
    manual_curp: Optional[CurpField]
    calculated_curp: CurpField
    validated_curp: Optional[CurpField]
    renapo_curp_match: bool
    renapo_full_match: bool

    @classmethod
    def create(
        cls,
        names: str,
        first_surname: str,
        date_of_birth: dt.date,
        country_of_birth: str,
        state_of_birth: State,
        gender: Gender,
        second_surname: Optional[str] = None,
        manual_curp: Optional[CurpField] = None,
        *,
        session: Session = global_session,
    ) -> 'CurpValidation':
        req = CurpValidationRequest(
            names=names,
            first_surname=first_surname,
            second_surname=second_surname,
            date_of_birth=date_of_birth,
            state_of_birth=state_of_birth,
            country_of_birth=country_of_birth,
            gender=gender,
            manual_curp=manual_curp,
        )
        return cast(
            'CurpValidation',
            cls._create(session=session, **req.dict()),
        )



import datetime as dt

import pytest

from cuenca import ApiKey


@pytest.mark.vcr
def test_api_keys_create():
    api_key = ApiKey.create()
    assert api_key.id is not None
    assert api_key.secret is not None
    assert api_key.active


@pytest.mark.vcr
def test_api_keys_retrieve():
    id_key = 'AKMPSxy2UeSKqU1J6spDNwqA'
    api_key: ApiKey = ApiKey.retrieve(id_key)
    assert api_key.id == id_key
    assert api_key.secret == '********'


@pytest.mark.vcr
def test_api_key_deactivate():
    api_key = ApiKey.create()
    assert api_key.active

    disabled = ApiKey.deactivate(api_key.id, 0)
    assert disabled.id == api_key.id
    assert disabled.deactivated_at is not None
    assert not disabled.active

    api_key.refresh()
    assert not api_key.active


@pytest.mark.vcr
def test_update_api_key():
    fields_to_update = dict(
        metadata=dict(info='v1.1.1'),
        user_id='USrZMoC3sDgEDteFNMYhcXY',
        platform_id='PTZbBlk__kQt-wfwzP5nwA9A',
    )
    api_key_id = 'AKNm37rN3JQsStxnI_L3n4Lw'

    updated = ApiKey.update(api_key_id, **fields_to_update)
    assert updated.user_id == fields_to_update['user_id']
    assert not hasattr(updated, 'metadata')
    assert not hasattr(updated, 'platform_id')


def test_api_key_to_dict():
    created = dt.datetime.now()
    date = created.astimezone(dt.timezone.utc).isoformat()
    api_key: ApiKey = ApiKey(
        id='12345',
        secret='********',
        created_at=created,
        deactivated_at=None,
        updated_at=date,
        user_id=None,
    )
    api_key_dict = dict(
        id='12345',
        secret='********',
        created_at=date,
        deactivated_at=None,
        updated_at=date,
        user_id=None,
    )
    assert api_key_dict == api_key.to_dict()


def test_api_key_from_dict():
    api_keys_dict = dict(
        id='123455',
        secret='*********',
        created_at=dt.datetime.utcnow(),
        deactivated_at=None,
        extra_field_1='not necessary',
        extra_field_2=12345,
        updated_at=dt.datetime.utcnow(),
        metadata=None,
        user_id=None,
    )
    api_key = ApiKey._from_dict(api_keys_dict)
    assert not hasattr(api_key, 'extra_field_1')
    assert not hasattr(api_key, 'extra_field_2')
    assert api_key.id is not None



from typing import ClassVar, cast

from cuenca_validations.types import CommissionType, EntryType
from pydantic.dataclasses import dataclass

from .base import Transaction
from .resources import retrieve_uri

mapper = {CommissionType.cash_deposit: EntryType.credit}


@dataclass
class Commission(Transaction):
    _resource: ClassVar = 'commissions'

    type: CommissionType
    related_transaction_uri: str

    @property  # type: ignore
    def related_transaction(self) -> Transaction:
        return cast(Transaction, retrieve_uri(self.related_transaction_uri))
